using System;
using System.IO;
using System.Collections;

namespace IFOInfo
{
    /// <summary>
    /// Summary description for IFOFile.
    /// </summary>
    public class IFOFile
    {
        private string m_sFileName;
        private string m_sErrors;
        private bool m_bValid = false;

        // private M_VOB m_M_VOB;
        // private VTS_ATRT m_ATRT;
        // private TXTDT_MGI m_TXTDT_MGI;
        // private M_LUI m_M_LUI;
        // private MAIT m__MAIT;

        private I_MAT m_I_MAT;
        private PTT_SRPTI m_PTT_SRPTI;
        private M_PGCI_UT m_M_PGCI_UT;
        private TMAPT m_TMAPT;
        private C_ADT m_M_C_ADT;
        private VOBU_ADMAP m_M_VOBU_ADMAP;
        private C_ADT m_C_ADT;
        private VOBU_ADMAP m_VOBU_ADMAP;
        private VTS_PGCIT m_PGCIT;

        private const int iSectorSize = 2048;
        private byte[] m_ifoData;

        public IFOFile(string sFile)
        {
            m_sFileName = sFile;
            m_sErrors = "";
            m_bValid = false;

            StreamReader reader = null;

            try
            {
                reader = new StreamReader(sFile);
            }
            catch(Exception ex)
            {
                AddError("The file \"{0}\" could not be opened for reading [{1}]", sFile, ex.Message);
                return;
            }

            BinaryReader ifoReader = new BinaryReader(reader.BaseStream);

            m_ifoData = ifoReader.ReadBytes((int) ifoReader.BaseStream.Length);

            ifoReader.Close();
            ifoReader = null;

            m_bValid = ParseSectors();
        }

        public struct ChapterList
        {
            public int ProgramChain;
            public string Chapters;
            public int iMaxTime;
            public int iNumChapters;
        }

        public ChapterList[] GetChapterList()
        {
            ChapterList[] chapters = new ChapterList[m_PGCIT.arrayPGCI_SRP.Length];

            for(int i = 0; i < m_PGCIT.arrayPGCI_SRP.Length; i++)
            {
                chapters[i].ProgramChain = i;
                chapters[i].iNumChapters = 0;

                string strOut = "";
                // strOut += string.Format("Program Chain {0}\n", i);
                strOut += string.Format("[Chapters]\n");
                PGC pgc = m_PGCIT.arrayPGCI_SRP[i].pgc;

                int iPrevTime = 0;
                int iMaxTime = 0;

                byte[] map = pgc.baPGCProgramMap;

                int[] iaCellTimes = new int[pgc.cellPlaybackIT.Length];
                for(int j = 0; j < iaCellTimes.Length; j++)
                {
                    int iTime = TimeInMS(pgc.cellPlaybackIT[j].iPlaybackTime);
                    if (pgc.cellPlaybackIT[j].iCategory < 0)
                    {
                        // This is probably an alternate angle, but I'm not sure what the
                        // "right" way to check this is.
                        continue;
                    }

                    iaCellTimes[j] = iPrevTime;

                    if (iTime > iMaxTime)
                        iMaxTime = iTime;

                    iPrevTime = iTime + iPrevTime;
                }

                iPrevTime = 0;
                for(int j = 0; j < map.Length; j++)
                {
                    int iTime = iaCellTimes[map[j]-1];
                    int iOrigTime = TimeInMS(pgc.cellPlaybackIT[map[j]-1].iPlaybackTime);
                    if ((j == map.Length - 1) && (iOrigTime < 5000))
                        break;

                    strOut += GetChapterFormat(j+1, iTime);
                    chapters[i].iNumChapters++;
                }

                if (iMaxTime >= 0x500)
                {
                    chapters[i].Chapters = strOut;
                }

                chapters[i].iMaxTime = iMaxTime;
            }

            return chapters;
        }

        private string GetChapterFormat(int iChapter, int iTime)
        {
            string strResult;

            switch(Format)
            {
                case ChapterFormat.Frames:
                    strResult = string.Format("{0}={1} Chapter {2}\n", iChapter, (int) (iTime * FPS / 1000 + 0.5), iChapter);
                    break;
                default:  // case ChapterFormat.HMS:
                    strResult = string.Format("{0}={1} Chapter {2}\n", iChapter, MSToHMS(iTime), iChapter);
                    break;
            }

            return strResult;
        }

        private int FromBCD(int iBCD)
        {
            // in BCD, 0x57 = 57
            int iResult = 0;
            int iMultiplier = 1;

            while (iBCD > 0)
            {
                iResult += iMultiplier * (iBCD & 0xf);
                iBCD >>= 4;
                iMultiplier *= 10;
            }

            return iResult;
        }

        private int TimeInMS(int iTime)
        {
            // Converts annoying BCD time format into millisecond-based format
            int iMS = FromBCD(iTime % 0x40) * 1000 / 30;
            iMS += FromBCD(iTime / 0x100 % 0x100) * 1000;
            iMS += FromBCD(iTime / 0x10000 % 0x100) * (60 * 1000);
            iMS += FromBCD(iTime / 0x1000000) * (60 * 60 * 1000);

            return iMS;
        }

        private string MSToHMS(int iMS)
        {
            return string.Format("{0:00}:{1:00}:{2:00}.{3:000}", iMS / (1000 * 60 * 60), iMS / (1000 * 60) % 60, iMS / 1000 % 60, iMS % 1000);
        }

        private void Parse_I_MAT(int iSector)
        {
            m_I_MAT.sVMGIdentifier = StringAt(iSector, 0, 12);
            m_I_MAT.iLastSectorVTS = IntAt(iSector, 12);
            m_I_MAT.iLastSectorVTSI = IntAt(iSector, 28);
            m_I_MAT.bSpecificationVersion = ByteAt(iSector, 33);
            m_I_MAT.iVTSCategory = IntAt(iSector, 34);
            m_I_MAT.iVMGNumVolumes = ShortAt(iSector, 38);
            m_I_MAT.iVMGThisVolume = ShortAt(iSector, 40);
            m_I_MAT.bDiscSide = ByteAt(iSector, 42);
            m_I_MAT.iVMGNumTitleSets = ShortAt(iSector, 62);
            m_I_MAT.baProviderID = BytesAt(iSector, 64, 32);         // 32 bytes
            m_I_MAT.baPOSCode = BytesAt(iSector, 96, 8);             // 8 bytes
            m_I_MAT.iVTSI_MAT_End = IntAt(iSector, 128);
            m_I_MAT.iFirstPGC_Start = IntAt(iSector, 132);

            m_I_MAT.iVTSM_VOBS_Start = IntAt(iSector, 192);
            m_I_MAT.iVTSTT_VOBS_Start = IntAt(iSector, 196);
            m_I_MAT.iVTS_PTT_SRPT_Start = IntAt(iSector, 200);
            m_I_MAT.iVTS_PGCIT_Start = IntAt(iSector, 204);
            m_I_MAT.iVTSM_PGCI_UT_Start = IntAt(iSector, 208);
            m_I_MAT.iVTS_TMAPT_Start = IntAt(iSector, 212);
            m_I_MAT.iVTSM_C_ADT_Start = IntAt(iSector, 216);
            m_I_MAT.iVTSM_VOBU_ADMAP_Start = IntAt(iSector, 220);
            m_I_MAT.iVTS_C_ADT_Start = IntAt(iSector, 224);
            m_I_MAT.iVTS_VOBU_ADMAP_Start = IntAt(iSector, 228);
            m_I_MAT.iVTSM_VOBS_VideoAttr = ShortAt(iSector, 256);

            m_I_MAT.bVTSMNumAudio = ByteAt(iSector, 259);
            m_I_MAT.baVTSMAudioAttr = Bytes2DAt(iSector, 260, m_I_MAT.bVTSMNumAudio, 8);     // Array of 8-byte attributes
            m_I_MAT.bVTSMNumSBStreams = ByteAt(iSector, 341);
            m_I_MAT.baVTSVTSMSubAttr = Bytes2DAt(iSector, 342, m_I_MAT.bVTSMNumSBStreams, 6);

            m_I_MAT.iVTSTT_VOBSVidAttr = ShortAt(iSector, 512);
            m_I_MAT.bVTSTT_VOBSNumAudio = ByteAt(iSector, 515);
            m_I_MAT.baVTSTT_VOBSAudAttr = Bytes2DAt(iSector, 516, m_I_MAT.bVTSTT_VOBSNumAudio, 8);

            m_I_MAT.bVTSTT_VOBSNumSub = ByteAt(iSector, 597);
            m_I_MAT.baVTSTT_VOBSSubAttr = Bytes2DAt(iSector, 598, m_I_MAT.bVTSTT_VOBSNumSub, 6);
        }

        private void AddError(string strFormat, params string[] arrStrings)
        {
            m_sErrors += String.Format(strFormat, arrStrings) + "\n";
        }

        public bool Valid
        {
            get
            {
                return m_bValid;
            }
        }

        public string Errors
        {
            get
            {
                return m_sErrors;
            }
        }

        private string StringAt(int iSector, int iStart, int iLength)
        {
            string sResult = "";
            iStart += (iSector * iSectorSize);

            for(int i = iStart; i < iStart + iLength; i++)
            {
                sResult += (char) m_ifoData[i];
            }

            return sResult;
        }

        private int IntAt(int iSector, int iOffset)
        {
            iOffset += (iSector * iSectorSize);

            // Assume big-endian
            return (m_ifoData[iOffset] << 24) | (m_ifoData[iOffset + 1] << 16) | (m_ifoData[iOffset + 2] << 8) | m_ifoData[iOffset + 3];
        }

        private byte[] BytesAt(int iSector, int iStart, int iLength)
        {
            iStart += (iSector * iSectorSize);

            byte[] baResult = new byte[iLength];
            for(int i = iStart, j = 0; i < iStart + iLength; i++, j++)
            {
                baResult[j] = m_ifoData[i];
            }

            return baResult;
        }

        private int[] IntsAt(int iSector, int iStart, int iCount)
        {
            int[] iaResult = new int[iCount];

            for(int i = iStart, j = 0; j < iCount; i += 4, j++)
            {
                iaResult[j] = IntAt(iSector, i);
            }

            return iaResult;
        }

        private short[] ShortsAt(int iSector, int iStart, int iCount)
        {
            short[] iaResult = new short[iCount];

            for(int i = iStart, j = 0; i < iStart + (iCount * 2); i += 2, j++)
            {
                iaResult[j] = ShortAt(iSector, i);
            }

            return iaResult;
        }

        private long[] LongsAt(int iSector, int iStart, int iCount)
        {
            long[] iaResult = new long[iCount];

            for(int i = iStart, j = 0; i < iStart + (iCount * 8); i += 8, j++)
            {
                iaResult[j] = (IntAt(iSector, i) << 32) | IntAt(iSector, i+4);
            }

            return iaResult;
        }

        private byte[,] Bytes2DAt(int iSector, int iStart, int iCount, int iInnerSize)
        {
            iStart += (iSector * iSectorSize);
            byte[,] baResult = new byte[iCount, iInnerSize];

            for (int i = iStart, j = 0; j < iCount; i += iInnerSize, j++)
            {
                for(int z = 0; z < iInnerSize; z++)
                {
                    baResult[j, z] = m_ifoData[i+z];
                }
            }

            return baResult;
        }

        private short ShortAt(int iSector, int iOffset)
        {
            iOffset += (iSector * iSectorSize);
            // Assume big-endian
            return (short) ((m_ifoData[iOffset] << 8) | m_ifoData[iOffset + 1]);
        }

        private byte ByteAt(int iSector, int iOffset)
        {
            return m_ifoData[iSector * iSectorSize + iOffset];
        }

        private bool ParseSectors()
        {
            try { Parse_I_MAT(0); }
            catch(Exception ex)
            {
                AddError("Invalid I_MAT format [{0}]", ex.Message);
                return false;
            }

            try { Parse_M_VOB(m_I_MAT.iVTSM_VOBS_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_VOB format [{0}]", ex.Message);
                return false;
            }

            try { Parse_PTT_SRPTI(m_I_MAT.iVTS_PTT_SRPT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid PTT_SRPTI format [{0}]", ex.Message);
                return false;
            }
            
            try { Parse_M_PGCI_UT(m_I_MAT.iVTSM_PGCI_UT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_PGCI_UT format [{0}]", ex.Message);
                return false;
            }

            try { Parse_PGCIT(m_I_MAT.iVTS_PGCIT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid VTS PGCIT format [{0}]", ex.Message);
                return false;
            }
            
            try { Parse_TMAPT(m_I_MAT.iVTS_TMAPT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid MAPTI format [{0}]", ex.Message);
                return false;
            }
            
            try { Parse_VTS_ATRT(m_I_MAT.iVS_ATRT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid VTS_ATRT format [{0}]", ex.Message);
                return false;
            }
            
            try { Parse_M_C_ADT(m_I_MAT.iVTSM_C_ADT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_C_ADT format [{0}]", ex.Message);
                return false;
            }

            try { Parse_M_VOBU_ADMAP(m_I_MAT.iVTSM_VOBU_ADMAP_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_VOBU_ADMAP format [{0}]", ex.Message);
                return false;
            }

            try { Parse_C_ADT(m_I_MAT.iVTS_C_ADT_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_C_ADT format [{0}]", ex.Message);
                return false;
            }

            try { Parse_VOBU_ADMAP(m_I_MAT.iVTS_VOBU_ADMAP_Start); }
            catch(Exception ex)
            {
                AddError("Invalid M_VOBU_ADMAP format [{0}]", ex.Message);
                return false;
            }

            return true;
        }

        public ChapterFormat Format;
        public double FPS;

        private void Parse_M_VOB(int iSector)
        {
            if (iSector == 0)
                return;
        }

        private void Parse_PTT_SRPTI(int iSector)
        {
            if (iSector == 0)
                return;

            m_PTT_SRPTI.iNum_TUSP_TTU = ShortAt(iSector, 0);
            m_PTT_SRPTI.iPTT_SRPT_End = IntAt(iSector, 4);
            m_PTT_SRPTI.iaTTU_Offsets = IntsAt(iSector, 8, m_PTT_SRPTI.iNum_TUSP_TTU);

            if (m_PTT_SRPTI.iaTTU_Offsets.Length < 1)
                return;

            m_PTT_SRPTI.arrayTTU = new TTU[m_PTT_SRPTI.iNum_TUSP_TTU];

            for(int iTTU = 0; iTTU < m_PTT_SRPTI.iNum_TUSP_TTU; iTTU++)
            {
                int iOffset = m_PTT_SRPTI.iaTTU_Offsets[iTTU];
                int iMax = m_PTT_SRPTI.iPTT_SRPT_End;
                if (m_PTT_SRPTI.iaTTU_Offsets.Length > iTTU + 1)
                {
                    iMax = m_PTT_SRPTI.iaTTU_Offsets[iTTU+1];
                }

                int iNumPTTs = ((iMax + 1) - iOffset) / 4;

                m_PTT_SRPTI.arrayTTU[iTTU].ptt = new PTT[iNumPTTs];

                for (int iPTT = 0, iIndex = 0; iPTT < iNumPTTs; iPTT++, iIndex += 4)
                {
                    m_PTT_SRPTI.arrayTTU[iTTU].ptt[iPTT].PGCN = ShortAt(iSector, iOffset + iIndex);
                    m_PTT_SRPTI.arrayTTU[iTTU].ptt[iPTT].PGN = ShortAt(iSector, iOffset + iIndex + 2);
                }

            }
        }

        private void Parse_M_PGCI_UT(int iSector)
        {
            if (iSector == 0)
                return;

            m_M_PGCI_UT.iNum_VTSM_PGCI_LU = ShortAt(iSector, 0);
            m_M_PGCI_UT.iVTSM_PGCI_End = IntAt(iSector, 4);

            m_M_PGCI_UT.arrayPGCILU = new VTSM_PGCI_LU[m_M_PGCI_UT.iNum_VTSM_PGCI_LU];
            for (int i = 8, j = 0; j < m_M_PGCI_UT.iNum_VTSM_PGCI_LU; j++, i += 8)
            {
                m_M_PGCI_UT.arrayPGCILU[j].iVTSMLC = ShortAt(iSector, i);
                m_M_PGCI_UT.arrayPGCILU[j].bVTSMEM = ByteAt(iSector, i + 3);
                m_M_PGCI_UT.arrayPGCILU[j].VTSM_LU_Start = IntAt(iSector, i + 4);
                m_M_PGCI_UT.arrayPGCILU[j].vtsm_lu = Parse_VTSM_LU(iSector, m_M_PGCI_UT.arrayPGCILU[j].VTSM_LU_Start);
            }
        }

        private VTSM_LU Parse_VTSM_LU(int iSector, int iOffset)
        {
            VTSM_LU vtsm_lu;

            vtsm_lu.iNumPGCI_SRP = ShortAt(iSector, iOffset);
            vtsm_lu.iVTSM_LU_EA_End = ShortAt(iSector, iOffset + 4);
            vtsm_lu.arrayPGCI_SRP = new PGCI_SRP[vtsm_lu.iNumPGCI_SRP];

            vtsm_lu.arrayPGCI_SRP = ParsePGCI_SRP(iSector, iOffset, vtsm_lu.iNumPGCI_SRP);

            return vtsm_lu;
        }

        private PGC ParsePGC(int iSector, int iOffset)
        {
            PGC pgc;

            pgc.bNumPrograms = ByteAt(iSector, iOffset+2);
            pgc.bNumCells = ByteAt(iSector, iOffset+3);
            pgc.iPlaybackTime = IntAt(iSector, iOffset+4);
            pgc.iPUOP = IntAt(iSector, iOffset+8);
            pgc.iaAudioStreamStatus = ShortsAt(iSector, iOffset+12, 8);
            pgc.iaSubPictureStreamStatus = IntsAt(iSector, iOffset+28, 32);
            pgc.iNextPGC = ShortAt(iSector, iOffset+156);
            pgc.iPrevPGC = ShortAt(iSector, iOffset+158);
            pgc.iUpPGC = ShortAt(iSector, iOffset+160);
            pgc.bStillTime = ByteAt(iSector, iOffset+162);
            pgc.bPGPlaybackMode = ByteAt(iSector, iOffset+163);
            pgc.iaColorYCrCB = IntsAt(iSector, iOffset+164, 16);
            pgc.iCommandTable_Start = ShortAt(iSector, iOffset+228);
            pgc.iProgramMap_Start = ShortAt(iSector, iOffset+230);
            pgc.iCellPlaybackIT_Start = ShortAt(iSector, iOffset+232);
            pgc.iCellPositionIT_Start = ShortAt(iSector, iOffset+234);

            pgc.pgcCommandTable = ParseCommandTable(iSector, pgc.iCommandTable_Start + iOffset);
            pgc.baPGCProgramMap = ParseProgramMap(iSector, pgc.iProgramMap_Start + iOffset, pgc.bNumPrograms);

            pgc.cellPlaybackIT = ParseCellPlaybackIT(iSector, pgc.iCellPlaybackIT_Start + iOffset, pgc.bNumCells);
            pgc.cellPositionIT = ParseCellPositionIT(iSector, pgc.iCellPositionIT_Start + iOffset, pgc.bNumCells);

            return pgc;
        }

        private CellPositionInfo[] ParseCellPositionIT(int iSector, int iOffset, int iNum)
        {
            if (iOffset == 0)
                return new CellPositionInfo[0];

            CellPositionInfo[] result = new CellPositionInfo[iNum];

            for(int i = iOffset, j = 0; i < (iOffset + iNum * 4); i += 4, j++)
            {
                result[j].iVOBID = ShortAt(iSector, i);
                result[j].bCellID = ByteAt(iSector, i + 3);
            }

            return result;
        }

        private CellPlaybackInfo[] ParseCellPlaybackIT(int iSector, int iOffset, int iNum)
        {
            if (iOffset == 0)
                return new CellPlaybackInfo[0];

            CellPlaybackInfo[] result = new CellPlaybackInfo[iNum];

            for(int i = iOffset, j = 0; i < (iOffset + iNum * 24); i += 24, j++)
            {
                result[j].iCategory = ShortAt(iSector, i);
                result[j].bStillTime = ByteAt(iSector, i+2);
                result[j].bCommand = ByteAt(iSector, i+3);
                result[j].iPlaybackTime = IntAt(iSector, i+4);
                result[j].iEntryPointSector = IntAt(iSector, i+8);
                result[j].iFirstILVU_VOBU_End = IntAt(iSector, i+12);
                result[j].iLastVOBU_Start = IntAt(iSector, i+16);
                result[j].iThisCellLastSector = IntAt(iSector, i+20);
            }

            return result;
        }

        private byte[] ParseProgramMap(int iSector, int iOffset, int iNum)
        {
            if (iOffset == 0)
                return new byte[0];

            byte[] baResult = new byte[iNum];

            for(int i = 0; i < iNum; i++)
            {
                baResult[i] = ByteAt(iSector, iOffset + i);
            }

            return baResult;
        }

        private PGCCommandTable NullCommandTable
        {
            get
            {
                PGCCommandTable table;

                table.iNumPreCommands = 0;
                table.iNumPostCommands = 0;
                table.iNumCellCommands = 0;
                table.laCellCommands = new long[0];
                table.laPostCommands = new long[0];
                table.laPreCommands = new long[0];

                return table;
            }
        }

        private PGCCommandTable ParseCommandTable(int iSector, int iOffset)
        {
            if (iOffset == 0)
                return NullCommandTable;

            PGCCommandTable table;

            table.iNumPreCommands = ShortAt(iSector, iOffset);
            table.iNumPostCommands = ShortAt(iSector, iOffset+2);
            table.iNumCellCommands = ShortAt(iSector, iOffset+4);
            table.laPreCommands = LongsAt(iSector, iOffset+8, table.iNumPreCommands);
            table.laPostCommands = LongsAt(iSector, iOffset + 8 + (8 * table.iNumPreCommands), table.iNumPostCommands);
            table.laCellCommands = LongsAt(iSector, iOffset + 8 + (8 * table.iNumPreCommands) + (8 * table.iNumPostCommands), table.iNumCellCommands);

            return table;
        }

        private PGCI_SRP[] ParsePGCI_SRP(int iSector, int iOffset, int iNum)
        {
            PGCI_SRP[] result = new PGCI_SRP[iNum];

            for(int i = iOffset + 8, j = 0; j < iNum; i += 8, j++)
            {
                result[j].bCatMenuMask = ByteAt(iSector, i);
                result[j].bCategory = ByteAt(iSector, i+1);
                result[j].iCatParentalMask = ShortAt(iSector, i+2);
                result[j].iPGCI_Start = IntAt(iSector, i+4);
                result[j].pgc = ParsePGC(iSector, result[j].iPGCI_Start + iOffset);
            }

            return result;
        }

        private void Parse_PGCIT(int iSector)
        {
            if (iSector == 0)
                return;

            m_PGCIT.iNum_PGCI_SRP = ShortAt(iSector, 0);
            m_PGCIT.iVTS_PGCI_End = IntAt(iSector, 4);
            m_PGCIT.arrayPGCI_SRP = ParsePGCI_SRP(iSector, 0, m_PGCIT.iNum_PGCI_SRP);
        }

        private void Parse_M_LUI(int iSector)
        {
            if (iSector == 0)
                return;
        }

        private void Parse_MAIT(int iSector)
        {
            if (iSector == 0)
                return;
        }

        private void Parse_TMAPT(int iSector)
        {
            if (iSector == 0)
                return;

            m_TMAPT.iNumTMAPS = ShortAt(iSector, 0);
            m_TMAPT.iTMAPS_End = IntAt(iSector, 4);
            m_TMAPT.iaTimeMap_Start = IntsAt(iSector, 8, m_TMAPT.iNumTMAPS);
            m_TMAPT.timeMap = new TimeMap[m_TMAPT.iNumTMAPS];

            for(int i = 0; i < m_TMAPT.iNumTMAPS; i++)
            {
                m_TMAPT.timeMap[i] = ParseTimeMap(iSector, m_TMAPT.iaTimeMap_Start[i]);
            }
        }

        private TimeMap ParseTimeMap(int iSector, int iOffset)
        {
            TimeMap map;

            map.bTimeUnit = ByteAt(iSector, iOffset);
            map.iNumMapEntries = ShortAt(iSector, iOffset + 2);
            map.iaMap = IntsAt(iSector, 4, map.iNumMapEntries);

            return map;
        }

        private void Parse_VTS_ATRT(int iSector)
        {
            if (iSector == 0)
                return;
        }

        private void Parse_TXTDT_MGI(int iSector)
        {
            if (iSector == 0)
                return;
        }

        private void Parse_M_C_ADT(int iSector)
        {
            if (iSector == 0)
                return;

            m_M_C_ADT.iNumVOBs = ShortAt(iSector, 0);
            m_M_C_ADT.iCADT_End = IntAt(iSector, 4);
            m_M_C_ADT.cellAddressTable = ParseCellAddressTable(iSector, 8, (m_M_C_ADT.iCADT_End - 8) / 12);
        }

        private CellAddress[] ParseCellAddressTable(int iSector, int iOffset, int iCount)
        {
            CellAddress[] result = new CellAddress[iCount];

            for(int i = iOffset, j = 0; j < iCount; i += 12, j++)
            {
                result[j].iVOBID = ShortAt(iSector, i);
                result[j].bCellID = ByteAt(iSector, i+2);
                result[j].iStartSector = IntAt(iSector, i+4);
                result[j].iEndSector = IntAt(iSector, i+8);
            }

            return result;
        }

        private void Parse_M_VOBU_ADMAP(int iSector)
        {
            if (iSector == 0)
                return;

            m_M_VOBU_ADMAP.iVOBU_ADMAP_End = IntAt(iSector, 0);
            m_M_VOBU_ADMAP.iaVOBU_Start = IntsAt(iSector, 4, (m_M_VOBU_ADMAP.iVOBU_ADMAP_End - 4) / 4);
        }
	
        private void Parse_C_ADT(int iSector)
        {
            if (iSector == 0)
                return;

            m_C_ADT.iNumVOBs = ShortAt(iSector, 0);
            m_C_ADT.iCADT_End = IntAt(iSector, 4);
            m_C_ADT.cellAddressTable = ParseCellAddressTable(iSector, 8, (m_C_ADT.iCADT_End - 8) / 12);
        }

        private void Parse_VOBU_ADMAP(int iSector)
        {
            if (iSector == 0)
                return;

            m_VOBU_ADMAP.iVOBU_ADMAP_End = IntAt(iSector, 0);
            m_VOBU_ADMAP.iaVOBU_Start = IntsAt(iSector, 4, (m_VOBU_ADMAP.iVOBU_ADMAP_End - 4) / 4);
        }
	
    }

    // Video Manager Information Management Table
    public struct I_MAT
    {
        public string sVMGIdentifier;
        public int iLastSectorVTS;
        public int iLastSectorVTSI;         // or VMGI
        public byte bSpecificationVersion;
        public int iVTSCategory;            // or VGM category
        public short iVMGNumVolumes;
        public short iVMGThisVolume;
        public byte bDiscSide;
        public short iVMGNumTitleSets;
        public byte[] baProviderID;         // 32 bytes
        public byte[] baPOSCode;            // 8 bytes
        public int iVTSI_MAT_End;
        public int iFirstPGC_Start;
        public int iVTSM_VOBS_Start;
        public int iVTSTT_VOBS_Start;
        public int iVTS_PTT_SRPT_Start;
        public int iVTSM_PGCI_UT_Start;
        public int iVTS_PGCIT_Start;
        public int iVS_ATRT_Start;
        public int iVTS_TMAPT_Start;
        public int iVTSM_C_ADT_Start;
        public int iVTSM_VOBU_ADMAP_Start;
        public int iVTS_C_ADT_Start;
        public int iVTS_VOBU_ADMAP_Start;
        public short iVTSM_VOBS_VideoAttr;

        public byte bVTSMNumAudio;
        public byte[,] baVTSMAudioAttr;     // Array of 8-byte attributes
        public byte bVTSMNumSBStreams;
        public byte[,] baVTSVTSMSubAttr;

        public short iVTSTT_VOBSVidAttr;
        public byte bVTSTT_VOBSNumAudio;
        public byte[,] baVTSTT_VOBSAudAttr;
        public byte bVTSTT_VOBSNumSub;
        public byte[,] baVTSTT_VOBSSubAttr;

    }

    // Video Manager Menu VideoObjectSet
    public struct M_VOB
    {
    }

    public struct PTT
    {
        public short PGCN;
        public short PGN;
    }

    public struct TTU
    {
        public PTT[] ptt;
    }

    // Part of title search pointer table information
    public struct PTT_SRPTI
    {
        public short iNum_TUSP_TTU;     // Number of Title Unit Search Pointers TTUs in the VTS
        public int iPTT_SRPT_End;       // End byte of PTT_SRPT table
        public int[] iaTTU_Offsets;
        public TTU[] arrayTTU;
    }

    public struct PGCCommandTable
    {
        public short iNumPreCommands;
        public short iNumPostCommands;
        public short iNumCellCommands;

        public long[] laPreCommands;
        public long[] laPostCommands;
        public long[] laCellCommands;
    }

    public struct CellPlaybackInfo
    {
        public short iCategory;
        public byte bStillTime;
        public byte bCommand;
        public int iPlaybackTime;
        public int iEntryPointSector;
        public int iFirstILVU_VOBU_End;
        public int iLastVOBU_Start;
        public int iThisCellLastSector;
    }

    public struct CellPositionInfo
    {
        public short iVOBID;
        public byte bCellID;
    }

    public struct PGC       // Program Chain
    {
        public byte bNumPrograms;
        public byte bNumCells;
        public int iPlaybackTime;
        public int iPUOP;
        public short[] iaAudioStreamStatus;
        public int[] iaSubPictureStreamStatus;
        public short iNextPGC;
        public short iPrevPGC;
        public short iUpPGC;
        public byte bStillTime;     // in seconds
        public byte bPGPlaybackMode;
        public int[] iaColorYCrCB;
        public short iCommandTable_Start;
        public short iProgramMap_Start;
        public short iCellPlaybackIT_Start;
        public short iCellPositionIT_Start;

        public PGCCommandTable pgcCommandTable;
        public byte[] baPGCProgramMap;
        public CellPlaybackInfo[] cellPlaybackIT;
        public CellPositionInfo[] cellPositionIT;
    }

    public struct PGCI_SRP
    {
        public byte bCatMenuMask;
        public byte bCategory;
        public short iCatParentalMask;
        public int iPGCI_Start;

        public PGC pgc;
    }

    public struct VTSM_LU   // Video Title Set Menu Language Unit
    {
        public short iNumPGCI_SRP;
        public int iVTSM_LU_EA_End;

        public PGCI_SRP[] arrayPGCI_SRP;
    }

    public struct VTSM_PGCI_LU
    {
        public short iVTSMLC;           // Video Title Set Menu Language Code
        public byte bVTSMEM;            // Video Title Set Menu Existence Mask
        public int VTSM_LU_Start;         // Video Title Set Menu Language start byte
        public VTSM_LU vtsm_lu;
    }

    // Video Manager Menu PGCI Unit Table
    public struct M_PGCI_UT
    {
        public short iNum_VTSM_PGCI_LU;     // Number of VTSM_PGCI_LUs
        public int iVTSM_PGCI_End;          // End byte of VTSM_PGCI_LUs table

        public VTSM_PGCI_LU[] arrayPGCILU;
    }

    // Video Title Set Program Chain Info Table
    public struct VTS_PGCIT
    {
        public short iNum_PGCI_SRP;
        public int iVTS_PGCI_End;
        
        public PGCI_SRP[] arrayPGCI_SRP;
    }

    // Video Manager Menu Language Unit Table
    public struct M_LUI
    {
    }

    // Parental Management Information Table
    public struct MAIT
    {
    }

    public struct TimeMap
    {
        public byte bTimeUnit;     // in seconds
        public short iNumMapEntries;
        public int[] iaMap;
    }

    // Video Title Set Time Map Table
    public struct TMAPT
    {
        public short iNumTMAPS;
        public int iTMAPS_End;
        public int[] iaTimeMap_Start;
        public TimeMap[] timeMap;
    }

    // Video Title Set Attribute Table
    public struct VTS_ATRT
    {
    }

    // Text Data Manager information
    public struct TXTDT_MGI
    {
    }

    public struct CellAddress
    {
        public short iVOBID;
        public byte bCellID;
        public int iStartSector;
        public int iEndSector;
    }

    // Cell Address Table information
    public struct C_ADT
    {
        public short iNumVOBs;
        public int iCADT_End;
        public CellAddress[] cellAddressTable;
    }

    // Video Title Set VOBU Address Map
    public struct VOBU_ADMAP
    {
        public int iVOBU_ADMAP_End;
        public int[] iaVOBU_Start;
    }

    public enum ChapterFormat
    {
        HMS,
        Frames
    }
}
